/*

  Copyright (c) 1998, 1999, Digital Creations, Fredericksburg, VA, USA.  
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are
  met:
  
    o Redistributions of source code must retain the above copyright
      notice, this list of conditions, and the disclaimer that follows.
  
    o Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions, and the following disclaimer in
      the documentation and/or other materials provided with the
      distribution.
  
    o Neither the name of Digital Creations nor the names of its
      contributors may be used to endorse or promote products derived
      from this software without specific prior written permission.
  
  
  THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS AND CONTRIBUTORS *AS
  IS* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
  TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL
  CREATIONS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
  USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
  DAMAGE.

  $Id: oci_.i,v 1.7 1999/08/20 13:48:13 petrilli Exp $

  If you have questions regarding this software,
  contact:
 
    Digital Creations L.C.  
    info@digicool.com
 
    (540) 371-6909

*/
%module oci_
%{

#define ODESSP_ASIZE 128

#include "ociapr.h"
#include "Buffer.h"

static void
PyVar_Assign(PyObject **v,  PyObject *e)
{
  Py_XDECREF(*v);
  *v=e;
}
#define ASSIGN(V,E) PyVar_Assign(&(V),(E))
#define UNLESS(E) if(!(E))
#define UNLESS_ASSIGN(V,E) ASSIGN(V,E); UNLESS(V)
#define OBJECT(O) ((PyObject*)(O))

static char *
PySequence_CharArray(PyObject *s, int *d)
{
  if((*d=PyString_Size(s)) < 0) return NULL;
  return PyString_AsString(s);
}  

static PyObject *ErrorObject;

#define I_TO_TUPLE(T) static PyObject * T ## _to_tuple(T *a, int l) { \
  PyObject *r, *e; \
  UNLESS(r=PyTuple_New(l)) return NULL; \
  while(--l >= 0) { \
      UNLESS(e=PyInt_FromLong(a[l])) { Py_DECREF(r); return NULL; } \
      PyTuple_SET_ITEM(r,l,e); \
    } \
  return r; \
} 

I_TO_TUPLE(ub2)
I_TO_TUPLE(ub1)
I_TO_TUPLE(ub4)
I_TO_TUPLE(sb2)

static PyObject *
OracleError(struct cda_def *lda, sb2 rcode)
  {
    text buf[512];
    sword l;
    PyObject *v;

    Py_BEGIN_ALLOW_THREADS
    l=oerhms(lda, rcode, buf, 511);
    Py_END_ALLOW_THREADS
    if((v=Py_BuildValue("is#",(int)rcode,buf,l)))
      {
	PyErr_SetObject(ErrorObject, v);
	Py_DECREF(v);
      }
    return NULL;
  }

typedef struct {
  PyObject_HEAD
  PyObject *open;
  struct cda_def lda;
  /* Hopefully, using long will do the right thing on 64-bit machines. */
  long hda[64];	
} Connection;

static PyObject *
ConnectionClosedError()
{
  PyErr_SetString(PyExc_TypeError,
		  "The Oracle Connection is closed!");
  return NULL;
}

static void
Connection_dealloc(Connection *self) {
  if(self->open)
    {
      if(self->open==Py_None) ologof(&(self->lda));
      Py_DECREF(self->open);
    }
  
  PyMem_DEL(self);
}

static PyTypeObject ConnectionType = {
  PyObject_HEAD_INIT(NULL) 0,
  "Connection",	/*tp_name*/ sizeof(Connection),	0,  /* basic and item size*/
  (destructor)Connection_dealloc,	/*tp_dealloc*/
  0,0,0,0,0, /* print, getattr, setattr, cmp, repr */
  0,0,0,     /* tp_as_number, sequence, mapping */
  0,0,0,0,0, /* hash, call, str, getattro, setattro */
  0 ,0 , "Oracle database connection"
};

typedef struct {
  PyObject_HEAD
  PyObject *connection;
  struct cda_def cda;  
} Cursor;

static void
Cursor_dealloc(Cursor *self) {
  oclose(&(self->cda));
  Py_DECREF(self->connection);
  PyMem_DEL(self);
}

static PyObject *
Cursor_getattro(Cursor *self, PyObject *n)
{
  char *name;

  if((name=PyString_AsString(n)) && *name=='l' && strcmp(name,"lda")==0)
    {
      Py_INCREF(self->connection);
      return self->connection;
    }
  PyErr_SetObject(PyExc_AttributeError, n);
  return NULL;
}

static PyTypeObject CursorType = {
  PyObject_HEAD_INIT(NULL) 0,
  "Cursor",  /* name*/ sizeof(Cursor), 0, /* basic and item size */
  (destructor)Cursor_dealloc,	/*tp_dealloc*/
  0,0,0,0,0, /* print, getattr, setattr, cmp, repr */
  0,0,0,     /* tp_as_number, sequence, mapping */
  0,0,0,     /* hash, call, str, */
  (getattrofunc)Cursor_getattro,0, /* getattro, setattro */
  0L,0L, "Oracle database cursor" 
};

#ifdef DCORACLE8
#include "oci_8.c"
#endif

%}

%include pointer.i
%include typemaps.i
%include Buffer.i
%extern oratypes.h
%include ocidfn.h

/***************************************************************************
  Text typemaps                                                            */

%typemap(python, in) text *IN(int string_dim) { 
    if($source==Py_None) { $target=NULL; string_dim=0; }
    else UNLESS($target=(text*)PySequence_CharArray($source, &string_dim))
	   return NULL;  
}
%typemap(python, ignore) int STRING_DIM {}
%typemap(python, check)  int STRING_DIM { $target=string_dim;}
/***************************************************************************/

%rename closeConnection ologof;
%inline %{

static PyObject *
Connect(text *IN, int STRING_DIM)
  {
    Connection *self;
    sword status;

    UNLESS(self = PyObject_NEW(Connection, &ConnectionType)) return NULL;
    self->open=NULL;
    memset(&(self->lda),0,sizeof(self->lda));
    memset(&(self->hda),0,sizeof(self->hda));

    Py_BEGIN_ALLOW_THREADS
    status=olog(&(self->lda),(ub1 *)&(self->hda), 
		IN, STRING_DIM, NULL, 0, NULL, 0, (ub4)OCI_LM_DEF);
    Py_END_ALLOW_THREADS

    if(status)
      {
	OracleError(&(self->lda), self->lda.rc);
	Py_DECREF(self);
	return NULL;
      }

    self->open=Py_None;
    Py_INCREF(self->open);
    return (PyObject *)self;
  }

static PyObject *
closeConnection(PyObject *connection)
  {
    UNLESS(connection->ob_type==&ConnectionType)
      {
	PyErr_SetString(PyExc_TypeError,
	"Expected an Oracle Connection object");
	return NULL;
      }

    UNLESS(((Connection*)connection)->open) return ConnectionClosedError();

    Py_BEGIN_ALLOW_THREADS
    if (((Connection*)connection)->open == Py_None)
      ologof(&(((Connection*)connection)->lda));
    Py_END_ALLOW_THREADS
    Py_DECREF(((Connection*)connection)->open);
    ((Connection*)connection)->open=0;
    Py_INCREF(Py_None);
    return Py_None;
  }

static PyObject *
newCursor(PyObject *connection)
  {
    Cursor *self;
    sword status;
    
    UNLESS(connection->ob_type==&ConnectionType)
      {
	PyErr_SetString(PyExc_TypeError,
	"Expected an Oracle Connection object");
	return NULL;
      }

    UNLESS(((Connection*)connection)->open) return ConnectionClosedError();

    UNLESS(self = PyObject_NEW(Cursor, &CursorType)) return NULL;
    self->connection=connection;
    Py_INCREF(connection);

    Py_BEGIN_ALLOW_THREADS
    status=oopen(&(self->cda),&(((Connection*)connection)->lda),
		 NULL,-1,-1,NULL,-1);
    Py_END_ALLOW_THREADS
    
    if(status)
      {
	OracleError(&(((Connection*)connection)->lda), self->cda.rc);
	Py_DECREF(self);
	return NULL;
      }

    return (PyObject *)self;
  }

%}

%typemap(python, in) struct cda_def *lda {
  if($source->ob_type != &ConnectionType) {
    PyErr_SetString(PyExc_TypeError, "database connection required");
    return NULL;
  }
  UNLESS(((Connection*)$source)->open) return ConnectionClosedError();
  $target=&(((Connection*)$source)->lda);
}

%typemap(python, in) struct cda_def *cursor {
  if($source->ob_type != &CursorType) {
    PyErr_SetString(PyExc_TypeError, "database cursor required");
    return NULL;
  }
  UNLESS(((Connection*)((Cursor*)$source)->connection)->open)
       return ConnectionClosedError();
  $target=&(((Cursor*)$source)->cda);
}

%inline %{

static PyObject *
OracleErrorMessage(struct cda_def *lda, sb2 rcode)
  {
    text buf[512];
    sword l;
    PyObject *v;

    Py_BEGIN_ALLOW_THREADS
    l=oerhms(lda, rcode, buf, 511);
    Py_END_ALLOW_THREADS
    return PyString_FromStringAndSize((char*)buf, l);
  }

/* Utility routines to extract the cda from a lda or cursor */
struct cda_def *dbcda(struct cda_def *lda) {return lda;}
struct cda_def *cucda(struct cda_def *cursor) {return cursor;}
  
%}

%apply text *IN {text *objnam};
%apply int STRING_DIM {int onlen};
%typemap(python, ignore) ub1 *rsv1, size_t rsv1ln, ub1 *rsv2, size_t rsv2ln {
  $target=NULL;
}
%rename odessp_ odessp;
%inline %{
  static PyObject *
    odessp_(struct cda_def *lda, text *objnam, int onlen)
    {
      /* Note that we count on the Python global interpreter lock to
       * to let us get away with these static declarations. 
       */
      static ub2 ovrld[ODESSP_ASIZE];
      static ub2 pos[ODESSP_ASIZE];
      static ub2 level[ODESSP_ASIZE];
      static text argnm[ODESSP_ASIZE][30];
      static ub2 arnlen[ODESSP_ASIZE];
      static ub2 dtype[ODESSP_ASIZE];
      static ub1 defsup[ODESSP_ASIZE];
      static ub1 mode[ODESSP_ASIZE];
      static ub4 dtsize[ODESSP_ASIZE];
      static sb2 prec[ODESSP_ASIZE];
      static sb2 scale[ODESSP_ASIZE];
      static ub1 radix[ODESSP_ASIZE];
      static ub4 spare[ODESSP_ASIZE];
      ub4 arrsiz = (ub4) ODESSP_ASIZE;
      PyObject *r, *e, *s;
      int i;
      sword status;

      Py_BEGIN_ALLOW_THREADS
      status=odessp(lda, objnam, onlen, NULL, 0, NULL, 0,
		    ovrld,  pos,   level,  (text**)argnm,
		    arnlen, dtype, defsup, mode,
		    dtsize,  prec,  scale,  radix,
		    spare,  &arrsiz);
      Py_END_ALLOW_THREADS

      if(status)
	{
	  OracleError(lda, lda->rc);
	  return NULL;
	}
      
      UNLESS(r=PyTuple_New(11)) return NULL;

      UNLESS(e=ub2_to_tuple( ovrld, (int)arrsiz)) goto err;
      PyTuple_SET_ITEM(r, 0, e);
      UNLESS(e=ub2_to_tuple(   pos, (int)arrsiz)) goto err;
      PyTuple_SET_ITEM(r, 1, e);
      UNLESS(e=ub2_to_tuple( level, (int)arrsiz)) goto err;
      PyTuple_SET_ITEM(r, 2, e);
      UNLESS(e=PyTuple_New((int)arrsiz)) goto err;
      PyTuple_SET_ITEM(r, 3, e);
      for(i=0; i < arrsiz; i++)
	{
	  UNLESS(s=PyString_FromStringAndSize((char*)argnm[i],arnlen[i]))
	    goto err;
	  PyTuple_SET_ITEM(e, i, s);
	}
      UNLESS(e=ub2_to_tuple( dtype, (int)arrsiz)) goto err;
      PyTuple_SET_ITEM(r, 4, e);
      UNLESS(e=ub1_to_tuple(defsup, (int)arrsiz)) goto err;
      PyTuple_SET_ITEM(r, 5, e);
      UNLESS(e=ub1_to_tuple(  mode, (int)arrsiz)) goto err;
      PyTuple_SET_ITEM(r, 6, e);
      UNLESS(e=ub4_to_tuple(dtsize, (int)arrsiz)) goto err;
      PyTuple_SET_ITEM(r, 7, e);
      UNLESS(e=sb2_to_tuple(  prec, (int)arrsiz)) goto err;
      PyTuple_SET_ITEM(r, 8, e);
      UNLESS(e=sb2_to_tuple( scale, (int)arrsiz)) goto err;
      PyTuple_SET_ITEM(r, 9, e);
      UNLESS(e=ub1_to_tuple( radix, (int)arrsiz)) goto err;
      PyTuple_SET_ITEM(r,10, e);

      return r;
    err:
      Py_DECREF(r);
      return NULL;
    }
%}

%except(python) {
   Py_BEGIN_ALLOW_THREADS
   $function
   Py_END_ALLOW_THREADS
}

%typemap(python, ignore) text *fmt {$target=NULL;}
%typemap(python, ignore) sb4 fmtl {$target=-1;}
%typemap(python, ignore) sword fmtt {$target=-1;}

/*
 Oci BIND (Piecewise or with Skips) 
 */
%apply text *IN {text *sqlvar};
%apply int STRING_DIM {sb4 sqlvl};
%apply void *BUFFER_OR_NULL { sb2 *indp };
%apply void *BUFFER_OR_NULL { ub2 *alen };
%apply void *BUFFER_OR_NULL { ub2 *arcode };
%apply void *BUFFER_OR_NULL { ub4 *cursiz };

sword  obindps(struct cda_def *cursor, ub1 opcode, text *sqlvar, 
	       sb4 sqlvl, ub1 *pvctx, sb4 progvl, 
	       sword ftype, sword scale,
	       sb2 *indp, ub2 *alen, ub2 *arcode, 
	       sb4 pv_skip, sb4 ind_skip, sb4 alen_skip, sb4 rc_skip,
	       ub4 maxsiz, ub4 *cursiz,
	       text *fmt, sb4 fmtl, sword fmtt);


%typemap(python, in) void *BUFFER_OR_WHATEVERO {
  if($source==Py_None) $target=NULL;
  else if($source->ob_type==BufferType)
    $target=(void*)(((Buffer*)$source)->val.c);
  else if(PyString_Check($source))
    $target=(void*)(((PyStringObject*)$source)->ob_sval);
  else if(PyFloat_Check($source))
    $target=(void*)&(PyFloat_AS_DOUBLE((PyFloatObject*)$source));
  else if(PyInt_Check($source))
    $target=(void*)&(PyInt_AS_LONG((PyIntObject*)$source));
  else if($source->ob_type==&CursorType)
    {
      UNLESS(((Connection*)((Cursor*)$source)->connection)->open)
	return ConnectionClosedError();
      $target=(void*)&(((Cursor*)$source)->cda);
    }
  else
    {
      PyErr_SetString(PyExc_TypeError, "buffer expected");
      return NULL;
    }
}
%apply void *BUFFER_OR_WHATEVERO { ub1 *pvctx };

%inline %{
  static sword bindpw(struct cda_def *cursor, text *sqlvar, sb4 sqlvl,
		      PyObject *pvctx, sb4 progvl, 
		      sword ftype, sword scale,
		      sb2 *indp, ub2 *arcode, 
		      ub4 maxsiz, ub4 *cursiz)
    {
      return obindps(cursor, 0, sqlvar, sqlvl, (ub1 *)pvctx, progvl,
		     ftype, 0,
		     indp, 0, arcode,
		     0,0,0,0,
		     maxsiz, cursiz, NULL, 0, 0);
    }

  static sword bind(struct cda_def *cursor, text *sqlvar, sb4 sqlvl,
		    ub1 *pvctx, sb4 progvl, sword ftype,
		    sb2 *indp, ub2 *alen, ub2 *arcode,
		    ub4 maxsiz, ub4 *cursiz)
    {
      return obindps(cursor, 1, sqlvar, sqlvl, pvctx, progvl,
		     ftype, 0, indp, alen, arcode, progvl, 2, 2, 2,
		     maxsiz, cursiz, NULL, 0, 0);
    }
		     
%}

sword  obreak(struct cda_def *lda);
sword  ocan  (struct cda_def *cursor);
sword  oclose(struct cda_def *cursor);
sword  ocof  (struct cda_def *lda);
sword  ocom  (struct cda_def *lda);
sword  ocon  (struct cda_def *lda);


/*
 * Oci DEFINe (Piecewise or with Skips) 
 */
%apply void *BUFFER { ub1 *bufctx };
%apply void *BUFFER { sb2 *rlen };
%apply void *BUFFER { sb2 *rcode };
sword  odefinps(struct cda_def *cursor, ub1 opcode, sword pos,
		ub1 *bufctx, sb4 bufl, sword ftype, sword scale, 
		sb2 *indp, text *fmt, sb4 fmtl, sword fmtt, 
		ub2 *alen, ub2 *arcode,
		sb4 pv_skip, sb4 ind_skip, sb4 alen_skip, sb4 rc_skip);

%apply void *BUFFER { ub1 *bufctx };

%inline %{
  static sword definpw(struct cda_def *cursor, sword pos, PyObject *bufctx,
		sb4 bufl, sword ftype, sword scale, 
		sb2 *indp, ub2 *alen, ub2 *arcode)
    {
      return odefinps(cursor, 0, pos, (ub1 *)bufctx, bufl,
		      ftype, 0,
		      indp, NULL, 0, 0, 
		      alen, arcode, 0,0,0,0);
    }
  static sword define(struct cda_def *cursor, sword pos,
		      ub1 *bufctx,
		      sb4 bufl, sword ftype, 
		      sb2 *indp, ub2 *alen, ub2 *arcode)
    {
      return odefinps(cursor, 1, pos, bufctx,
		bufl, ftype, 0, 
		indp, NULL, 0, 0,
		alen, arcode,
		bufl, 2, 2, 2);
    }
%}

%apply int *T_OUTPUT { sb4 *dbsize };
%apply short *T_OUTPUT { sb2 *dbtype };
%apply int *OUTPUT_STRING_DIM {sb4 *cbufl };
%apply int *T_OUTPUT { sb4 *dsize };
%apply short *T_OUTPUT { sb2 *prec };
%apply short *T_OUTPUT { sb2 *scale };
%apply short *T_OUTPUT { sb2 *nullok };
%typemap(python, ignore) sb1 *cbuf (sb1 buf1[256]) {
  $target=buf1;
  *$target=0;
}
%typemap(python, argout) sb1 *cbuf {
  $target = t_output_helper($target,
     *buf1
     ? PyString_FromStringAndSize((char*)buf1, output_string_dim1)
     : PyString_FromString("") 
     );
}
%typemap(python, ignore) sb4 *cbufl(long output_string_dim1) {
  $target=&output_string_dim1; output_string_dim1=256; 
} 
sword  odescr(struct cda_def *cursor, sword pos, sb4 *dbsize,
                 sb2 *dbtype, sb1 *cbuf, sb4 *cbufl, sb4 *dsize,
                 sb2 *prec, sb2 *scale, sb2 *nullok);
%clear sb1 *buf ;

sword  oexec    (struct cda_def *cursor);
sword  oexfet   (struct cda_def *cursor, ub4 nrows,
                 sword cancel, sword exact);
sword  oexn     (struct cda_def *cursor, sword iters, sword rowoff);
sword  ofen     (struct cda_def *cursor, sword nrows);
sword  ofetch   (struct cda_def *cursor);

%apply void *SIZED_BUFFER { ub1 *buf };
%apply int BUFFER_SIZE {sb4 bufl};
%apply int *T_OUTPUT { sb4 *retl };
sword  oflng    (struct cda_def *cursor, sword pos, ub1 *buf,
                 sb4 bufl, sword dtype, ub4 *retl, sb4 offset);


%apply unsigned char *T_OUTPUT { ub1 *piecep };
%apply unsigned int *T_OUTPUT { ub4 *iterp };
%apply unsigned int *T_OUTPUT { ub4 *indexp };
%typemap(python, ignore) dvoid **ctxpp (dvoid *temp) {$target=&temp; }
%typemap(python, argout) dvoid **ctxpp {
  Py_INCREF((PyObject*)*$source);
  $target = t_output_helper($target, (PyObject*)*$source);
}
sword  ogetpi   (struct cda_def *cursor, ub1 *piecep, dvoid **ctxpp, 
                 ub4 *iterp, ub4 *indexp);

sword  oopt     (struct cda_def *cursor, sword rbopt, sword waitopt);
sword  opinit   (ub4 mode);

%apply text *IN {text *sqlstm};
%apply int STRING_DIM {sb4 sqllen};
sword  oparse   (struct cda_def *cursor, text *sqlstm, sb4 sqllen,
                 sword defflg, ub4 lngflg);

sword  orol     (struct cda_def *lda);

%apply void *BUFFER { dvoid *bufp };
%apply void *BUFFER { ub4 *lenp };
sword  osetpi   (struct cda_def *cursor, ub1 piece, dvoid *bufp, ub4 *lenp);

/* non-blocking functions */
sword onbset    (struct cda_def *lda ); 
sword onbtst    (struct cda_def *lda ); 
sword onbclr    (struct cda_def *lda ); 



%init %{
  ConnectionType.ob_type=&PyType_Type;
  CursorType.ob_type=&PyType_Type;
  ErrorObject=PyString_FromString("OracleError");
  PyDict_SetItemString(d,"error",ErrorObject);
  PyDict_SetItemString(d,"ConnectionType",((PyObject*)&ConnectionType));
  PyDict_SetItemString(d,"CursorType",((PyObject*)&CursorType));

  PyDict_SetItemString(d,"__version__",
		       PyString_FromStringAndSize(
                          "$Revision: 1.7 $"+11,
                          strlen("$Revision: 1.7 $"+11)-2));

  PyDict_SetItemString(d,"_id",
     PyString_FromString("$Id: oci_.i,v 1.7 1999/08/20 13:48:13 petrilli Exp $"));

#ifdef DCORACLE8
  initoci_8();
#endif

%}













