
#include "oci.h"

static OCIEnv *envhp;

static PyObject *
RaiseOCIError(dvoid *hp)
{
  text errbuf[256];
  sb4 errcode;
  ub4 handle_type;

  Py_BEGIN_ALLOW_THREADS

  if (!hp) hp=envhp;
  if (hp==(dvoid *)envhp) handle_type=OCI_HTYPE_ENV;
  else                    handle_type=OCI_HTYPE_ERROR;

  OCIErrorGet (hp, 1, NULL, &errcode,
	       errbuf, sizeof(errbuf), handle_type);

  Py_END_ALLOW_THREADS

  PyErr_SetObject(ErrorObject,
		  Py_BuildValue("si", errbuf, (int)errcode));
  return NULL;
}

/* Declarations for objects of type ServerContext */

typedef struct {
  PyObject_HEAD
  OCISvcCtx *hp;
  OCIError *ep;
} ServerContext;

staticforward PyTypeObject ServerContextType;

typedef struct {
  PyObject_HEAD
  Connection *db;
  OCILobLocator *hp;
} Lob;
staticforward PyTypeObject LobType;

/* CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC */

static PyObject *
ServerContext_lda(ServerContext *self, PyObject *args)
{
  Connection *db;
  sword status;

  UNLESS(PyArg_ParseTuple(args,"")) return NULL;

  UNLESS(db = PyObject_NEW(Connection, &ConnectionType)) return NULL;
  db->open=OBJECT(self);
  Py_INCREF(self);
  memset(&(db->lda),0,sizeof(db->lda));
  memset(&(db->hda),0,sizeof(db->hda));

  Py_BEGIN_ALLOW_THREADS
  status=OCISvcCtxToLda (self->hp, self->ep, &db->lda);
  Py_END_ALLOW_THREADS
  if (status==OCI_ERROR)
    {
      Py_DECREF(db);
      return RaiseOCIError(self->ep);
    }

  return OBJECT(db);
}

static struct PyMethodDef ServerContext_methods[] = {
  {"lda", (PyCFunction)ServerContext_lda, METH_VARARGS,
   "lda(errorHandle) -- Get a OCI 7 connection"},
  {NULL,		NULL}		/* sentinel */
};

static PyObject *
Logon(PyObject *ignored, PyObject *args)
{
  ServerContext *self;
  char *u, *p, *d;
  int lu, lp, ld;
  sword status;
	
  UNLESS(PyArg_ParseTuple(args, "s#s#s#",
			  &u, &lu, &p, &lp, &d, &ld))
    return NULL;

  UNLESS(self = PyObject_NEW(ServerContext, &ServerContextType)) return NULL;

  Py_BEGIN_ALLOW_THREADS
  status=OCIHandleAlloc(envhp, (dvoid**)&self->ep, OCI_HTYPE_ERROR, NULL, NULL);
  Py_END_ALLOW_THREADS

  if(status==OCI_ERROR)
    {
      self->hp=NULL;
      self->ep=NULL;
      Py_DECREF(self);
      return RaiseOCIError(envhp);
    }

  Py_BEGIN_ALLOW_THREADS
  status=OCILogon(envhp, self->ep, &self->hp,
		  (text*)u, lu, (text*)p, lp, (text*)d, ld);
  Py_END_ALLOW_THREADS
  
  if(status==OCI_ERROR)
    {
      RaiseOCIError(self->ep);
      self->hp=NULL;
      Py_DECREF(self);
      return NULL;
    }

  return OBJECT(self);
}

static void
ServerContext_dealloc(ServerContext *self)
{
  Py_BEGIN_ALLOW_THREADS
  if (self->hp) OCILogoff(self->hp, self->ep);
  if (self->ep) OCIHandleFree(self->ep, OCI_HTYPE_ERROR);
  Py_END_ALLOW_THREADS
  PyMem_DEL(self);
}

static PyObject *
ServerContext_getattr(ServerContext *self, char *cname)
{
  return Py_FindMethod(ServerContext_methods, (PyObject *)self, cname);
}

static char ServerContexttype__doc__[] = 
"An Oracle 8 database connection"
;

static PyTypeObject ServerContextType = {
  PyObject_HEAD_INIT(NULL)
  0,				/*ob_size*/
  "ServerContext",			/*tp_name*/
  sizeof(ServerContext),		/*tp_basicsize*/
  0,				/*tp_itemsize*/
  /* methods */
  (destructor)ServerContext_dealloc,	/*tp_dealloc*/
  (printfunc)0,	/*tp_print*/
  (getattrfunc)ServerContext_getattr,		/*obsolete tp_getattr*/
  (setattrfunc)0,		/*obsolete tp_setattr*/
  (cmpfunc)0,	/*tp_compare*/
  (reprfunc)0,		/*tp_repr*/
  0,		/*tp_as_number*/
  0,		/*tp_as_sequence*/
  0,		/*tp_as_mapping*/
  (hashfunc)0,		/*tp_hash*/
  (ternaryfunc)0,	/*tp_call*/
  (reprfunc)0,		/*tp_str*/
  (getattrofunc)0,	/*tp_getattro*/
  (setattrofunc)0,	/*tp_setattro*/
  
  /* Space for future expansion */
  0L,0L,
  ServerContexttype__doc__ /* Documentation string */
};

/* LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL */

static PyObject *
Lob_define(Lob *self, PyObject *args)
{
  Cursor *c;
  Buffer *indp, *alen, *arcode;
  int pos, ftype;
  ServerContext *sc;
  sword status;

  UNLESS(PyArg_ParseTuple(args,"O!iiO!O!O!",
			  &CursorType, &c, &pos, &ftype,
			  BufferType, &indp, BufferType, &alen, BufferType, &arcode
			  )) return NULL;

  Py_BEGIN_ALLOW_THREADS
  status=odefinps(&c->cda, 1, pos, (ub1*)&self->hp,
		  sizeof(OCILobLocator *), ftype, 0, 
		  indp->val.h, NULL, 0, 0,
		  alen->val.H, arcode->val.H,
		  sizeof(OCILobLocator *), 2, 2, 2);
  Py_END_ALLOW_THREADS

  return PyInt_FromLong(status);
}

static PyObject *
Lob_write(Lob *self, PyObject *args)
{
  ub4 amtp;
  char *s;
  int l, csid=0, csfrm=SQLCS_IMPLICIT, offset=1;
  sword status;
  PyObject *r;
  ServerContext *sc;

  UNLESS(PyArg_ParseTuple(args,"s#|iii", &s, &l, &offset, &csid, &csfrm)) return NULL;
  amtp=l;
	
  if (! self->db->open || self->db->open->ob_type != &ServerContextType)
    {
      PyErr_SetString(ErrorObject,
		      "Database connection is clodes or not based on a\n"
		      "OCI 8 server context");
      return NULL;
    }
  sc=(ServerContext*)(self->db->open);

  Py_BEGIN_ALLOW_THREADS
  status=OCILdaToSvcCtx(&sc->hp, sc->ep, &self->db->lda);
  Py_END_ALLOW_THREADS
  if(status==OCI_ERROR) return RaiseOCIError(sc->ep);

  Py_BEGIN_ALLOW_THREADS
  status=OCILobWrite(sc->hp, sc->ep, self->hp, &amtp, offset, s, l,
		     OCI_ONE_PIECE, NULL, NULL, csid, csfrm);
  Py_END_ALLOW_THREADS
  if(status==OCI_ERROR) {
    RaiseOCIError(sc->ep);
    r=NULL;
  }
  else r=PyInt_FromLong(status);

  Py_BEGIN_ALLOW_THREADS
  status=OCISvcCtxToLda(sc->hp, sc->ep, &self->db->lda);
  Py_END_ALLOW_THREADS
  if(status==OCI_ERROR) return RaiseOCIError(sc->ep);

  return r;
}
    
static PyObject *
Lob_read(Lob *self, PyObject *args)
{
  ub4 amtp, ll;
  int l=-1, csid=0, csfrm=SQLCS_IMPLICIT, offset=1;
  sword status;
  PyObject *r=NULL;
  char *s;
  ServerContext *sc;
  
  UNLESS(PyArg_ParseTuple(args,"|iiii", &l, &offset, &csid, &csfrm)) return NULL;
	
  if (! self->db->open || self->db->open->ob_type != &ServerContextType)
    {
      PyErr_SetString(ErrorObject,
		      "Database connection is clodes or not based on a\n"
		      "OCI 8 server context");
      return NULL;
    }
  sc=(ServerContext*)(self->db->open);

  Py_BEGIN_ALLOW_THREADS
  status=OCILdaToSvcCtx(&sc->hp, sc->ep, &self->db->lda);
  Py_END_ALLOW_THREADS
  if(status==OCI_ERROR) return RaiseOCIError(sc->ep);

  Py_BEGIN_ALLOW_THREADS
  status=OCILobGetLength(sc->hp, sc->ep, self->hp, &ll);
  Py_END_ALLOW_THREADS
  if(status==OCI_ERROR)
    {
      RaiseOCIError(sc->ep);
      goto finally;
    }
  ll+=offset-1;
  if (l <= 0 || l > ll) l=ll;

  if (l < 1) return PyString_FromString("");

  amtp=l;
  r=PyString_FromStringAndSize(NULL, l);
  s=PyString_AsString(r);
  Py_BEGIN_ALLOW_THREADS
  status=OCILobRead(sc->hp, sc->ep, self->hp, &amtp, offset, s, l,
		    NULL, NULL, csid, csfrm);
  Py_END_ALLOW_THREADS
  if(status==OCI_ERROR) {
    RaiseOCIError(sc->ep);
    Py_DECREF(r);
    r=NULL;
  }

finally:
  Py_BEGIN_ALLOW_THREADS
  status=OCISvcCtxToLda(sc->hp, sc->ep, &self->db->lda);
  Py_END_ALLOW_THREADS
  if(status==OCI_ERROR) return RaiseOCIError(sc->ep);

  return r;
}
    
static PyObject *
Lob_length(Lob *self, PyObject *args)
{
  ub4 ll;
  sword status;
  PyObject *r=NULL;
  char *s;
  ServerContext *sc;
  
  UNLESS(PyArg_ParseTuple(args,"")) return NULL;
	
  if (! self->db->open || self->db->open->ob_type != &ServerContextType)
    {
      PyErr_SetString(ErrorObject,
		      "Database connection is clodes or not based on a\n"
		      "OCI 8 server context");
      return NULL;
    }
  sc=(ServerContext*)(self->db->open);

  Py_BEGIN_ALLOW_THREADS
  status=OCILdaToSvcCtx(&sc->hp, sc->ep, &self->db->lda);
  Py_END_ALLOW_THREADS
  if(status==OCI_ERROR) return RaiseOCIError(sc->ep);

  Py_BEGIN_ALLOW_THREADS
  status=OCILobGetLength(sc->hp, sc->ep, self->hp, &ll);
  Py_END_ALLOW_THREADS
  if(status==OCI_ERROR)
    {
      RaiseOCIError(sc->ep);
      goto finally;
    }
  r=PyInt_FromLong(ll);

finally:
  Py_BEGIN_ALLOW_THREADS
  status=OCISvcCtxToLda(sc->hp, sc->ep, &self->db->lda);
  Py_END_ALLOW_THREADS
  if(status==OCI_ERROR) return RaiseOCIError(sc->ep);

  return r;
}

static struct PyMethodDef Lob_methods[] = {
  {"length", (PyCFunction)Lob_length, METH_VARARGS,
   "length(errorHandle) -- Get the LOB length"},
  {"read", (PyCFunction)Lob_read, METH_VARARGS,
   "read(errorHandle, [n, offset, csid, csfrm) -- "
   "read from the LOB\n\n"
   "read n (or all) bytes at some offset (from 1) using\n"
   "a character set id(0) and character set form (SQLCS_IMPLICIT)\n"
  },
  {"write", (PyCFunction)Lob_write, METH_VARARGS,
   "read(errorHandle, s, [offset, csid, csfrm) -- "
   "write a string, s, to the LOB\n\n"
   "read n (or all) bytes at some offset (from 1) using\n"
   "a character set id(0) and character set form (SQLCS_IMPLICIT)\n"
  },
  {"define", (PyCFunction)Lob_define, METH_VARARGS,
   "define(cursor, position, type, ind, alen, arcode) -- "
   "define an output binding"
  },
  {NULL,		NULL}		/* sentinel */
};

static void
Lob_dealloc(Lob *self)
{
  if(self->hp)
    {
      Py_BEGIN_ALLOW_THREADS
      OCIDescriptorFree(self->hp, OCI_DTYPE_LOB);
      Py_END_ALLOW_THREADS
    }
  PyMem_DEL(self);
}

static PyObject *
Lob_getattr(Lob *self, char *cname)
{
  return Py_FindMethod(Lob_methods, (PyObject *)self, cname);
}

static char Lobtype__doc__[] = 
"Oracle Lob";

static PyTypeObject LobType = {
  PyObject_HEAD_INIT(NULL)
  0,				/*ob_size*/
  "Lob",			/*tp_name*/
  sizeof(Lob),		/*tp_basicsize*/
  0,				/*tp_itemsize*/
  /* methods */
  (destructor)Lob_dealloc,	/*tp_dealloc*/
  (printfunc)0,	/*tp_print*/
  (getattrfunc)Lob_getattr,		/*obsolete tp_getattr*/
  (setattrfunc)0,		/*obsolete tp_setattr*/
  (cmpfunc)0,	/*tp_compare*/
  (reprfunc)0,		/*tp_repr*/
  0,		/*tp_as_number*/
  0,		/*tp_as_sequence*/
  0,		/*tp_as_mapping*/
  (hashfunc)0,		/*tp_hash*/
  (ternaryfunc)0,	/*tp_call*/
  (reprfunc)0,		/*tp_str*/
  (getattrofunc)0,	/*tp_getattro*/
  (setattrofunc)0,	/*tp_setattro*/
  
  /* Space for future expansion */
  0L,0L,
  Lobtype__doc__ /* Documentation string */
};

/* MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM */

static PyObject *
newLob(PyObject *self, PyObject *args)
{
  Connection *db;
  Lob *lob;
  ServerContext *sc;
  sword status;

  UNLESS(PyArg_ParseTuple(args,"O!", &ConnectionType, &db)) return NULL;
	
  if (! db->open || db->open->ob_type != &ServerContextType)
    {
      PyErr_SetString(ErrorObject,
		      "Database connection is clodes or not based on a\n"
		      "OCI 8 server context");
      return NULL;
    }
  sc=(ServerContext*)(db->open);

  Py_BEGIN_ALLOW_THREADS
  status=OCILdaToSvcCtx(&sc->hp, sc->ep, &db->lda);
  Py_END_ALLOW_THREADS
  if(status==OCI_ERROR) return RaiseOCIError(sc->ep);

  UNLESS(lob = PyObject_NEW(Lob, &LobType)) return NULL;
  lob->db=db;
  Py_INCREF(db);
  Py_BEGIN_ALLOW_THREADS
  status=OCIDescriptorAlloc(envhp, (dvoid**)&lob->hp, OCI_DTYPE_LOB, NULL, NULL);
  Py_END_ALLOW_THREADS
  if(status==OCI_ERROR)
    {
      lob->hp=NULL;
      Py_DECREF(lob);
      return RaiseOCIError(envhp);
    }

  Py_BEGIN_ALLOW_THREADS
  status=OCISvcCtxToLda(sc->hp, sc->ep, &db->lda);
  Py_END_ALLOW_THREADS
  if(status==OCI_ERROR)
    {
      Py_DECREF(lob);
      return RaiseOCIError(sc->ep);
    }

  return OBJECT(lob);
}

static struct PyMethodDef Module_Level__methods[] = {
  {"lob", (PyCFunction)newLob, METH_VARARGS,
   "allocate a Lob descriptor"},
  {"Logon", (PyCFunction)Logon, METH_VARARGS,
   "Logon to the database"},
  {NULL, (PyCFunction)NULL, 0, NULL}		/* sentinel */
};

void
initoci_8()
{
  PyObject *m, *d;
  char *rev="$Revision: 1.3 $";

  /* Initialize Oracle */
  OCIInitialize(OCI_THREADED, NULL, NULL, NULL, NULL);
  if(OCIEnvInit(&envhp, OCI_DEFAULT, NULL, NULL)==OCI_ERROR)
    {
      RaiseOCIError(envhp);
      return;
    }

  /* Create the module and add the functions */
  m = Py_InitModule4("oci_8", Module_Level__methods,
		     "hack module to get access to OCI 8 LOBs",
		     (PyObject*)NULL,PYTHON_API_VERSION);

  /* Add some symbolic constants to the module */
  d = PyModule_GetDict(m);


  ServerContextType.ob_type=&PyType_Type;
  PyDict_SetItemString(d,"ServerContextType", OBJECT(&ServerContextType));

  LobType.ob_type=&PyType_Type;
  PyDict_SetItemString(d,"LobType", OBJECT(&LobType));

  PyDict_SetItemString(d, "error", ErrorObject);

  PyDict_SetItemString(d, "__version__",
		       PyString_FromStringAndSize(rev+11,strlen(rev+11)-2));
  
	
  /* Check for errors */
  if (PyErr_Occurred())
    Py_FatalError("can't initialize module oci_8");
}
